From 090d54e42e4a0efc5033e4443663021cb9db6f5f Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Hugo=20Beauz=C3=A9e-Luyssen?= <beauze.h@gmail.com>
Date: Tue, 24 Jan 2012 00:21:12 +0100
Subject: [PATCH 08/11] bluray: Support overlay updating & clearing.

---
 modules/access/bluray.c |  242 ++++++++++++++++++++++++++++++++++++++++++-----
 1 files changed, 219 insertions(+), 23 deletions(-)

diff --git a/modules/access/bluray.c b/modules/access/bluray.c
index 1f78975..7774ebc 100644
--- a/modules/access/bluray.c
+++ b/modules/access/bluray.c
@@ -33,8 +33,7 @@
 #include <vlc_demux.h>                      /* demux_t */
 #include <vlc_input.h>                      /* Seekpoints, chapters */
 #include <vlc_dialog.h>                     /* BD+/AACS warnings */
-#include <vlc_picture.h>
-#include <vlc_vout.h>
+#include <vlc_vout.h>                       /* vout_PutSubpicture / subpicture_t */
 
 #include <libbluray/bluray.h>
 #include <libbluray/keys.h>
@@ -67,9 +66,17 @@ vlc_module_begin ()
     set_callbacks( blurayOpen, blurayClose )
 vlc_module_end ()
 
-//libbluray's overlay.h defines 2 types of overlay.
+/* libbluray's overlay.h defines 2 types of overlay (bd_overlay_plane_e). */
 #define MAX_OVERLAY 2
 
+typedef enum    OverlayStatus
+{
+    Closed = 0,
+    ToDisplay,  //Used to mark the overlay to be displayed the first time.
+    Displayed,
+    Outdated    //used to update the overlay after it has been sent to the vout
+}               OverlayStatus;
+
 struct demux_sys_t
 {
     BLURAY         *bluray;
@@ -80,7 +87,10 @@ struct demux_sys_t
     input_title_t **pp_title;
 
     /* Overlay stuff */
-    subpicture_t    *p_pic[MAX_OVERLAY];
+    vlc_mutex_t             overlay_lock;
+    subpicture_t            *p_pic[MAX_OVERLAY];
+    OverlayStatus           overlay_status[MAX_OVERLAY];
+    subpicture_region_t     *p_regions[MAX_OVERLAY];
     int             current_overlay; // -1 if no current overlay;
 
     input_thread_t  *p_input;
@@ -90,6 +100,12 @@ struct demux_sys_t
     stream_t       *p_parser;
 };
 
+struct subpicture_updater_sys_t
+{
+    demux_sys_t *p_demux_sys;
+    int         i_plane;
+};
+
 /*****************************************************************************
  * Local prototypes
  *****************************************************************************/
@@ -131,11 +147,12 @@ static int blurayOpen( vlc_object_t *object )
     }
 
     /* */
-    p_demux->p_sys = p_sys = malloc(sizeof(*p_sys));
-    if (unlikely(!p_sys)) {
+    p_demux->p_sys = p_sys = calloc( 1, sizeof( *p_sys ) );
+    if ( unlikely( !p_sys ) )
+    {
         return VLC_ENOMEM;
     }
-    memset( p_sys, 0, sizeof( *p_sys ) );
+    vlc_mutex_init( &p_sys->overlay_lock );
     p_sys->current_overlay = -1;
 
     /* init demux info fields */
@@ -274,7 +291,7 @@ static void blurayClose( vlc_object_t *object )
         vlc_object_release( p_sys->p_input );
     if (p_sys->p_parser)
         stream_Delete(p_sys->p_parser);
-
+    vlc_mutex_destroy( &p_sys->overlay_lock );
     /* Titles */
     for (unsigned int i = 0; i < p_sys->i_title; i++)
         vlc_input_title_Delete(p_sys->pp_title[i]);
@@ -284,20 +301,119 @@ static void blurayClose( vlc_object_t *object )
 }
 
 /*****************************************************************************
- * libbluray overlay handling:
+ * subpicture_updater_t functions:
+ *****************************************************************************/
+
+static int      subpictureUpdaterValidate( subpicture_t *p_subpic,
+                                           bool b_fmt_src, const video_format_t *p_fmt_src,
+                                           bool b_fmt_dst, const video_format_t *p_fmt_dst,
+                                           mtime_t i_ts )
+{
+    VLC_UNUSED( b_fmt_src ); VLC_UNUSED( b_fmt_dst );
+    VLC_UNUSED( p_fmt_src ); VLC_UNUSED( p_fmt_dst );
+    VLC_UNUSED( i_ts );
+    subpicture_updater_sys_t    *p_upd_sys = p_subpic->updater.p_sys;
+    demux_sys_t                 *p_sys = p_upd_sys->p_demux_sys;
+
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    int res = p_sys->overlay_status[p_upd_sys->i_plane] == Outdated;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+    return res;
+}
+
+/* This should probably be moved to subpictures.c afterward */
+static subpicture_region_t*     subpicture_region_Clone( subpicture_region_t *p_region_src )
+{
+    if ( p_region_src == NULL )
+        return NULL;
+    subpicture_region_t *p_region_dst = subpicture_region_New( &p_region_src->fmt );
+    if ( unlikely( p_region_dst == NULL ) )
+        return NULL;
+    p_region_dst->i_x = p_region_src->i_x;
+    p_region_dst->i_y = p_region_src->i_y;
+    p_region_dst->i_align = p_region_src->i_align;
+    p_region_dst->i_alpha = p_region_src->i_alpha;
+
+    p_region_dst->psz_text = p_region_src->psz_text ? strdup( p_region_src->psz_text ) : NULL;
+    p_region_dst->psz_html = p_region_src->psz_html ? strdup( p_region_src->psz_html ) : NULL;
+    if ( p_region_src->p_style != NULL )
+    {
+        p_region_dst->p_style = malloc( sizeof( *p_region_dst->p_style ) );
+        p_region_dst->p_style = text_style_Copy( p_region_dst->p_style,
+                                                 p_region_src->p_style );
+    }
+
+    //Palette is already copied by subpicture_region_New, we just have to duplicate p_pixels
+    for ( int i = 0; i < p_region_src->p_picture->i_planes; i++ )
+        memcpy( p_region_dst->p_picture->p[i].p_pixels,
+                p_region_src->p_picture->p[i].p_pixels,
+                p_region_src->p_picture->p[i].i_lines * p_region_src->p_picture->p[i].i_pitch );
+    return p_region_dst;
+}
+
+static void     subpictureUpdaterUpdate( subpicture_t *p_subpic,
+                              const video_format_t *p_fmt_src,
+                              const video_format_t *p_fmt_dst,
+                              mtime_t i_ts )
+{
+    VLC_UNUSED( p_fmt_src ); VLC_UNUSED( p_fmt_dst ); VLC_UNUSED( i_ts );
+    subpicture_updater_sys_t    *p_upd_sys = p_subpic->updater.p_sys;
+    demux_sys_t                 *p_sys = p_upd_sys->p_demux_sys;
+
+    /*
+     * When this function is called, all p_subpic regions are gone.
+     * We need to duplicate our regions (stored internaly) to this subpic.
+     */
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    p_subpic->p_region = subpicture_region_Clone( p_sys->p_regions[p_upd_sys->i_plane] );
+    subpicture_region_t *p_src = p_sys->p_regions[p_upd_sys->i_plane]->p_next;
+    subpicture_region_t *p_dst = p_subpic->p_region;
+    while ( p_src != NULL )
+    {
+        p_dst->p_next = subpicture_region_Clone( p_src );
+        if ( p_dst->p_next == NULL )
+            break ;
+        p_dst = p_dst->p_next;
+        p_src = p_src->p_next;
+    }
+    p_dst->p_next = NULL;
+    p_sys->overlay_status[p_upd_sys->i_plane] = Displayed;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+}
+
+static void     subpictureUpdaterDestroy( subpicture_t *p_subpic )
+{
+    free( p_subpic->updater.p_sys );
+}
+
+/*****************************************************************************
+ * libbluray overlay callbacks:
  *****************************************************************************/
 
 static void blurayDeleteOverlaySubPic( demux_t *p_demux, int i_plane )
 {
     demux_sys_t     *p_sys = p_demux->p_sys;
 
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    p_sys->overlay_status[i_plane] = Closed;
     if ( p_sys->p_pic[i_plane] == NULL )
+    {
+        vlc_mutex_unlock( &p_sys->overlay_lock );
         return ;
+    }
+    //If we have a vout, let it cleanup by itself.
     if ( p_sys->p_vout )
         vout_FlushSubpictureChannel( p_sys->p_vout, p_sys->p_pic[i_plane]->i_channel );
     else
+    {
+        /*
+         * Else, the subpicture hasn't been sent yet, clean it ourselves.
+         * In that case, the vout doesn't hold the picture, we can use it safely
+         */
         subpicture_Delete( p_sys->p_pic[i_plane] );
+    }
     p_sys->p_pic[i_plane] = NULL;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
 }
 
 static void blurayCloseOverlay( demux_t *p_demux )
@@ -319,12 +435,28 @@ static void blurayCloseOverlay( demux_t *p_demux )
 static void blurayActivateOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
 {
     demux_sys_t     *p_sys = p_demux->p_sys;
+
+    /*
+     * If the overlay is already displayed, mark the picture as outdated.
+     * We must NOT use vout_PutSubpicture if a picture is already displayed.
+     */
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    if ( ( p_sys->overlay_status[ov->plane] == Displayed ||
+           p_sys->overlay_status[ov->plane] == Outdated )
+         && p_sys->p_vout )
+    {
+        p_sys->overlay_status[ov->plane] = Outdated;
+        vlc_mutex_unlock( &p_sys->overlay_lock );
+        return ;
+    }
     /*
      * Mark the overlay as available, but don't display it right now.
      * the blurayDemuxMenu will send it to vout, as it may be unavailable when
      * the overlay is computed
      */
     p_sys->current_overlay = ov->plane;
+    p_sys->overlay_status[ov->plane] = ToDisplay;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
 }
 
 static void blurayInitOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
@@ -334,7 +466,19 @@ static void blurayInitOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
     if ( p_sys->p_pic[ov->plane] != NULL )
         blurayDeleteOverlaySubPic( p_demux, ov->plane );
 
-    p_sys->p_pic[ov->plane] = subpicture_New( NULL );
+    subpicture_updater_sys_t    *p_upd_sys = malloc( sizeof( *p_upd_sys ) );
+    if ( unlikely( p_upd_sys == NULL ) )
+        return ;
+    p_upd_sys->p_demux_sys = p_sys;
+    p_upd_sys->i_plane = ov->plane;
+    subpicture_updater_t    updater =
+    {
+        .pf_validate = subpictureUpdaterValidate,
+        .pf_update = subpictureUpdaterUpdate,
+        .pf_destroy = subpictureUpdaterDestroy,
+        .p_sys = p_upd_sys,
+    };
+    p_sys->p_pic[ov->plane] = subpicture_New( &updater );
     p_sys->p_pic[ov->plane]->i_original_picture_width = ov->w;
     p_sys->p_pic[ov->plane]->i_original_picture_height = ov->h;
     p_sys->p_pic[ov->plane]->b_ephemer = true;
@@ -345,15 +489,25 @@ static void blurayClearOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
 {
     demux_sys_t     *p_sys = p_demux->p_sys;
 
-    //FIXME
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    subpicture_region_ChainDelete( p_sys->p_regions[ov->plane] );
+    p_sys->p_regions[ov->plane] = NULL;
+    p_sys->overlay_status[ov->plane] = Outdated;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
 }
 
 static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
 {
     demux_sys_t     *p_sys = p_demux->p_sys;
 
+    /*
+     * Compute a subpicture_region_t.
+     * It will be copied and sent to the vout later.
+     */
     if ( ov->img )
     {
+        vlc_mutex_lock( &p_sys->overlay_lock );
+
         video_format_t          fmt;
         video_format_Init( &fmt, 0 );
         video_format_Setup( &fmt, VLC_CODEC_YUVP, ov->w, ov->h, 1, 1 );
@@ -386,16 +540,54 @@ static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
             }
         }
         //Append the new region at the end of our current regions:
-        subpicture_region_t *p_reg = p_sys->p_pic[ov->plane]->p_region;
+        //If there's no region yet, don't bother checking for update...
+        subpicture_region_t *p_reg = p_sys->p_regions[ov->plane];
         if ( p_reg == NULL )
         {
-            p_sys->p_pic[ov->plane]->p_region = region;
+            p_sys->p_regions[ov->plane] = region;
+            vlc_mutex_unlock( &p_sys->overlay_lock );
             return ;
         }
-        while ( p_reg && p_reg->p_next )
-            p_reg = p_reg->p_next;
-        p_reg->p_next = region;
+        /*
+         * While adding the region, check if any region matches the one we're drawing.
+         * if so, delete the old region, and add the new one at the back of the list.
+         */
+        subpicture_region_t *p_prev = NULL;
+        while ( p_reg )
+        {
+            if ( p_reg->i_x == region->i_x && p_reg->i_y == region->i_y &&
+                 p_reg->fmt.i_width == region->fmt.i_width &&
+                 p_reg->fmt.i_height == region->fmt.i_height )
+            {
+                if ( p_prev == NULL )
+                {
+                    p_sys->p_regions[ov->plane] = p_reg->p_next;
+                    subpicture_region_Delete( p_reg );
+                    p_reg = p_sys->p_regions[ov->plane];
+                }
+                else
+                {
+                    p_prev->p_next = p_reg->p_next;
+                    subpicture_region_Delete( p_reg );
+                    p_reg = p_prev->p_next;
+                }
+            }
+            else
+            {
+                p_prev = p_reg;
+                p_reg = p_reg->p_next;
+            }
+        }
+        /* If there's only one region, both prev and reg will be NULL */
+        if ( p_prev == NULL )
+            p_sys->p_regions[ov->plane] = region;
+        else
+            p_prev->p_next = region;
+        vlc_mutex_unlock( &p_sys->overlay_lock );
     }
+    /*
+     * /!\ The region is now stored in our internal list, but not in the subpicture /!\
+     */
 }
 
 static void blurayOverlayProc( void *ptr, const BD_OVERLAY *const overlay )
@@ -411,13 +603,13 @@ static void blurayOverlayProc( void *ptr, const BD_OVERLAY *const overlay )
     switch ( overlay->cmd )
     {
         case BD_OVERLAY_INIT:
+            msg_Info( p_demux, "Initializing overlay" );
             blurayInitOverlay( p_demux, overlay );
             break ;
         case BD_OVERLAY_CLEAR:
             blurayClearOverlay( p_demux, overlay );
             break ;
         case BD_OVERLAY_FLUSH:
-            msg_Info( p_demux, "Displaying overlay" );
             blurayActivateOverlay( p_demux, overlay );
             break ;
         case BD_OVERLAY_DRAW:
@@ -457,7 +649,7 @@ static int  onMouseEvent( vlc_object_t *p_vout, const char *psz_var, vlc_value_t
     return VLC_SUCCESS;
 }
 
-static void blurayDisplayOverlay( demux_t *p_demux )
+static void bluraySendOverlayToVout( demux_t *p_demux )
 {
     demux_sys_t *p_sys = p_demux->p_sys;
 
@@ -467,7 +659,8 @@ static void blurayDisplayOverlay( demux_t *p_demux )
     p_sys->p_pic[p_sys->current_overlay]->i_channel =
             vout_RegisterSubpictureChannel( p_sys->p_vout );
     vout_PutSubpicture( p_sys->p_vout, p_sys->p_pic[p_sys->current_overlay] );
-    p_sys->current_overlay = -1;
+    //Mark the picture as updated, as it contains no region for now.
+    p_sys->overlay_status[p_sys->current_overlay] = Outdated;
     //Activating vout events handling:
     var_AddCallback( p_sys->p_vout, "mouse-moved", &onMouseEvent, p_demux );
     var_AddCallback( p_sys->p_vout, "mouse-clicked", &onMouseEvent, p_demux );
@@ -784,15 +977,18 @@ static int blurayDemuxMenu( demux_t *p_demux )
     }
     p_block->i_buffer = nread;
 
-    if ( p_sys->current_overlay != -1 )
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    if ( p_sys->current_overlay != -1 &&
+         p_sys->overlay_status[p_sys->current_overlay] == ToDisplay )
     {
+        vlc_mutex_unlock( &p_sys->overlay_lock );
         if ( p_sys->p_vout == NULL )
             p_sys->p_vout = input_GetVout( p_sys->p_input );
         if ( p_sys->p_vout != NULL )
-        {
-            blurayDisplayOverlay( p_demux );
-        }
+            bluraySendOverlayToVout( p_demux );
     }
+    else
+        vlc_mutex_unlock( &p_sys->overlay_lock );
     stream_DemuxSend( p_sys->p_parser, p_block );
     return 1;
 }
-- 
1.7.8.4

